#!/usr/bin/env python
#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
# 
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

"""
Build a de bruijn graph and split it into connected components, retaining
only those that are larger than a given threshold. Pickle the graph to a
file.
"""

import argparse
import networkx as nx
import os
import psutil
import re
import sys
import time

from Bio.SeqIO.QualityIO import FastqGeneralIterator

parser = argparse.ArgumentParser(
			formatter_class=argparse.RawDescriptionHelpFormatter,
			epilog=__doc__)
parser.add_argument('-n', type=int, metavar='N', default=10000, help="""
	only retain connected components with size >= N""")
parser.add_argument('K', type=int, help="k-mer size")
parser.add_argument('FASTQ', nargs='+', help="input reads")
parser.add_argument('--tips', action='store_true', help="keep tips")
args = parser.parse_args()

start_time = time.time()
proc = psutil.Process(os.getpid())

_graphml_header = """<?xml version="1.0" encoding="UTF-8"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
<key id="seq" for="edge" attr.name="seq" attr.type="string"/>"""

def log(msg):
	print "[%10.2f] %s" % (time.time() - start_time, msg)

def done():
	mem = proc.get_memory_info()
	log("done. / mem: {:12,d}KB / vmem: {:12,d}KB /".format(*mem))

alphabet = re.compile(r"^[ATCG]+$")
complement = { 'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C' }
G = nx.MultiDiGraph()

log("Hashing kmers and building node list...")

nreads = 0
ndiscard = 0

for fq in args.FASTQ:
	for _, seq, _ in FastqGeneralIterator(open(fq)):
		nreads += 1
		seq = str(seq)
		if alphabet.match(seq):
			for i in range(0, len(seq)-args.K+1):
				kmer = seq[i:i+args.K]
				G.add_node(kmer)
				rev = ''.join(complement[x] for x in reversed(kmer))
				G.add_node(rev)
		else:
			ndiscard += 1

print "processed %d reads:" % nreads
print "  discarded %d reads with Ns" % ndiscard
print "  kept %d unique kmers" % G.number_of_nodes()

done()

log("Building edge list...")

for node in G.nodes_iter():
	for base in ('A','C','G','T'):
		neighbor = base + node[:-1]
		if node != neighbor and G.has_node(neighbor):
			G.add_edge(neighbor, node, key=node[-1])
		neighbor = node[1:] + base
		if node != neighbor and G.has_node(neighbor):
			G.add_edge(node, neighbor, key=base)

print "%d edges" % G.number_of_edges()

done()

#######################
log("Removing tips...")
#######################

if not args.tips:

	ntails = 0
	nheads = 0
	tails = []
	heads = []
	for n in G.nodes_iter():
		if G.in_degree(n) == 0:
			tails.append(n)
		elif G.out_degree(n) == 0:
			heads.append(n)
	while tails:
		n = tails.pop()
		if G.has_node(n) and G.in_degree(n) == 0:
			tails += G.successors(n)
			G.remove_node(n)
			ntails += 1
	while heads:
		n = heads.pop()
		if G.has_node(n) and G.out_degree(n) == 0:
			heads += G.predecessors(n)
			G.remove_node(n)
			nheads += 1
	for n in G.nodes_iter():
		assert G.in_degree(n) > 0
		assert G.out_degree(n) > 0

	print "removed %d forward tips" % ntails
	print "removed %d reverse tips" % nheads
	print "%d nodes / %d edges" % (G.number_of_nodes(), G.number_of_edges())

	done()

#################################
log("Merging redundant nodes...")
#################################

nred = 0
for n in G.nodes():
	if G.in_degree(n) == 1 and G.out_degree(n) == 1:
		e1 = G.in_edges(n, keys=True)[0]
		e2 = G.out_edges(n, keys=True)[0]
		G.add_edge(e1[0], e2[1], key=e1[2]+e2[2])
		G.remove_node(n)
		nred += 1
for n in G.nodes():
	if G.in_degree(n) == 1:
		e1 = G.in_edges(n, keys=True)[0]
		for e in G.out_edges(n, keys=True):
			G.add_edge(e1[0], e[1], key=e1[2]+e[2])
		G.remove_node(n)
		nred += 1

print "merged %d redundant nodes" % nred
print "%d nodes / %d edges" % (G.number_of_nodes(), G.number_of_edges())

done()

##########################
log("Finding clusters...")
##########################

nclusters = 0
nvisited = 0

for n in G.nodes_iter():
	G.node[n]['visited'] = False

for n in G.nodes_iter():
	if not G.node[n]['visited']:
		queue = [n]
		cluster = set()
		# Follow all edges.
		while queue:
			n = queue.pop()
			if not G.node[n]['visited']:
				nvisited += 1
				if not nvisited % 10000:
					log("%d nodes visited..." % nvisited)
				cluster.add(n)
				queue += G.successors(n)
				queue += G.predecessors(n)
				G.node[n]['visited'] = True
		# Only retain large clusters.

		if len(cluster) > 1:
			print "cluster %d has %d nodes" % (nclusters, len(cluster))
			# Create a subgraph for the cluster.
			H = nx.MultiDiGraph()
			nedges = 1
			for n0, n1, k in G.edges_iter(keys=True):
				G[n0][n1][k]['visited'] = False
			for i, n in enumerate(cluster, start=1):
				G.node[n]['id'] = i
				H.add_node(i, kmer=n)
			for n in cluster:
				for n0, n1, k in G.out_edges(n, keys=True):
					if not G[n0][n1][k]['visited']:
						G[n0][n1][k]['visited'] = True
						H.add_edge(G.node[n0]['id'], G.node[n1]['id'], key=k, e=nedges)
						nedges += 1
			nx.write_graphml(H, 'cluster%d.graphml' % nclusters)
			nclusters += 1

done()

