#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
#
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import math
import os
import scipy.stats
import subprocess

from Bio import SeqIO
from random import random

from biolite import utils
from biolite.config import get_resource


class Sample:

	def __init__(self, likelihood, prior, length, ncontigs):
		self.posterior = likelihood + prior
		self.likelihood = likelihood
		self.prior = prior
		self.length = length
		self.ncontigs = ncontigs

	def odds_ratio(self, current):
		"""
		"""
		return (not math.isinf(self.likelihood)) and \
		       ((self.posterior - current.posterior) > math.log(random()))


class Sampler:

	def __init__(self, reads, length, ncontigs, scale):
		self.reads = reads
		self.nreads = int(subprocess.check_output(['wc', '-l', self.reads[0]]).split()[0])/4
		# Create a prior distribution from the estimated genome size.
		scale = float(scale)
		self.prior_length = scipy.stats.gamma(scale, scale=length/scale)
		self.prior_ncontigs = scipy.stats.gamma(scale, scale=ncontigs/scale)
		self.threads = get_resource('threads')

	def sample(self, contigs, filename):
		"""
		Run CGAL wrapper script to calculate log-likelihood of the contigs
		given the reads.
		"""
		# Write out contigs to FASTA file.
		length = 0
		ncontigs = 0
		with open(filename, 'w') as f:
			for contig in contigs:
				SeqIO.write(contig, f, 'fasta')
				length += len(contig.seq)
				ncontigs += 1
		with open('cgal.log', 'w') as f:
			try:
				likelihood = subprocess.check_output([
					'gabi-cgal-wrapper.sh', self.reads[0], self.reads[1],
					filename, self.threads],
					stderr=f)
			except subprocess.CalledProcessError as e:
				utils.die("failed to call gabi-cgal-wrapper.sh: check cgal.log\n", e)
		return Sample(
				float(likelihood) / self.nreads,
				math.log(self.prior_length.pdf(length) * self.prior_ncontigs.pdf(ncontigs)),
				length, ncontigs)

